/**********************************/
/*                                */
/*     db.c                       */
/*                                */
/*     database and schema        */
/*     code                       */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "mem.h"

#include "db.h"
#include "gen_io.h"

char *null_str = "?";

void new_schema( schema *sch, int size )
{
    int i;

    sch->size  = size;
    sch->elems = (attribute *)mem_malloc(size*sizeof(attribute));
    for ( i=0; i<size; i++ ) {
        sch->elems[i].name = NULL;
        sch->elems[i].type = ET_none;
        sch->elems[i].size = 0;
        }
}

schema *mk_new_schema( int size )
{
    schema *sch = (schema*)mem_malloc(sizeof(schema));
    new_schema(sch,size);
    return sch;
}

schema *copy_schema( schema *old_sch )
{
    int i;
    schema *sch = (schema*)mem_malloc(sizeof(schema));
    sch->size  = old_sch->size;
    sch->elems = (attribute *)mem_malloc(sch->size*sizeof(attribute));
    for ( i=0; i<sch->size; i++ ) {
        sch->elems[i] = old_sch->elems[i];
                             /* N.B. don't malloc new string */
        }
    return sch;
}

int equal_schema( schema *s1, schema *s2 )
{
    int i;
    if ( s1->size  !=  s2->size ) return 0;
    for ( i=0; i<s1->size; i++ ) {
        if ( s1->elems[i].type != s2->elems[i].type ) return 0;
        if ( s1->elems[i].size != s2->elems[i].size ) return 0;
        if ( strcmp(s1->elems[i].name, s2->elems[i].name ) != 0 ) return 0;
        }
    return 1;
}

void add_attribute_to_schema( schema *sch, attribute attrib  )
                                     /* updates existing schema */
                                     /* assumes malloced schema */
{
    int i;
    attribute *elems = sch->elems;
    sch->elems = (attribute *)mem_malloc((sch->size+1)*sizeof(attribute));
    for ( i=0; i<sch->size; i++ ) {
        sch->elems[i] = elems[i];
                                /* N.B. doesn't malloc new string */
        }
    sch->elems[i] = attrib;     /*   ditto   */
    sch->size++;
    mem_free(elems);
}

void add_schema_to_schema( schema *sch, schema *new  )
                                     /* updates existing schema */
                                     /* assumes malloced schema */
{
    int i;
    attribute *elems = sch->elems;
    sch->elems = (attribute *)mem_malloc((sch->size+new->size)*sizeof(attribute));
    for ( i=0; i<sch->size; i++ ) {
        sch->elems[i] = elems[i];
                                /* N.B. doesn't malloc new string */
        }
    for ( i=0; i<new->size; i++ ) {
        sch->elems[sch->size+i] = new->elems[i];
                                /* N.B. doesn't malloc new string */
        }
    sch->size += new->size;
    mem_free(elems);
}

void set_attribute_in_schema( schema *sch, attr_no attr,
                              char * name, elem_type type, int  size )
{
    assert( 0 <= attr && attr < sch->size );
    sch->elems[attr].name = name;
    sch->elems[attr].type = type;
    sch->elems[attr].size = size;
}


tuple mk_tuple( int size )
{
    return (tuple)mem_malloc( size * sizeof(tuple_elem) );
}

void empty_table( table *tb, schema *sch )
          /* consumes reference to schema, copy if necessary */
{
    tb->sch  = sch;
    tb->size = 0;
    tb->tups = NULL;
    tb->last = &tb->tups;
}

table *mk_empty_table( schema *sch )
          /* consumes reference to schema, copy if necessary */
{
    table *tb = (table*)mem_malloc(sizeof(table));
    empty_table(tb,sch);
    return tb;
}

void free_table( table *tb )
      /* does NOT free individual tuples */
      /* does NOT free schema            */
{
    tuple_list *tpl, *next;
    for ( tpl=tb->tups, next=tpl->next;
                           tpl != NULL;
                           tpl=next, next=next->next ) {
        mem_free(tpl);
        }
    mem_free(tb);
}

void free_tuples_and_table( table *tb )
      /* does NOT free schema            */
{
    tuple_list *tpl, *next;
    for ( tpl=tb->tups, next=tpl->next;
                           tpl != NULL;
                           tpl=next, next=next->next ) {
        mem_free(tpl->elem);
        mem_free(tpl);
        }
    mem_free(tb);
}

table *copy_table( table *tb )
      /* does NOT free schema            */
{
    table *new = mk_empty_table( tb->sch );
    tuple_list *otl, *next, **last;
    new->size = tb->size;
    last = &new->tups;;
    for ( otl=tb->tups; otl; otl=otl->next ) {
        next = (tuple_list*)mem_malloc(sizeof(tuple_list));
        *last = next;
        last = &next->next;
        next->elem = otl->elem;
        }
    *last = NULL;
    new->last = last;
    return new;
}

void add_to_table( table *tb, tuple tup )
     /* assumes tuple is appropriate for schema */
{
    tuple_list *new_tup = (tuple_list*)mem_malloc(sizeof(tuple_list));

    new_tup->next = NULL;
    new_tup->elem = tup;

    tb->size ++;

    *(tb->last) = new_tup;
    tb->last = &new_tup->next;
}

void add_cur_to_table( table *tb, currency *cur )
     /* assumes currency is appropriate for schema */
{
    add_to_table( tb, tuple_at_currency( cur ) );
}

void add_to_front_table( table *tb, tuple tup )
     /* assumes tuple is appropriate for schema */
{
    tuple_list *new_tup = (tuple_list*)mem_malloc(sizeof(tuple_list));

    new_tup->next = tb->tups;
    new_tup->elem = tup;

    tb->tups = new_tup;

    tb->size ++;

    if ( new_tup->next == NULL )   /*  adding to empty table  */
           tb->last = &new_tup->next;
}

void add_cur_to_front_table( table *tb, currency *cur )
     /* assumes currency is appropriate for schema */
{
    add_to_front_table( tb, tuple_at_currency( cur ) );
}

void append_table( table *front, table *back )
     /* updates front             */
     /* assumes tables compatable */
{
    table *new_back = copy_table(back);
    *(front->last) = new_back->tups;
    front->last = new_back->last;
    front->size += back->size;
    mem_free(new_back);
}

void prepend_table( table *front, table *back )
     /* updates back              */
     /* assumes tables compatable */
{
    table *new_front = copy_table(front);
    *(new_front->last) = back->tups;
    back->tups = new_front->tups;
    back->size += front->size;
    mem_free(new_front);
}

table *concat_table( table *front, table *back )
     /* updates neither           */
     /* assumes tables compatable */
{
    table *new_front = copy_table(front);
    append_table(new_front,back);
    return new_front;
}

table *add_attrib_to_table( table *tb, attribute attrib, tuple_elem init )
               /* mallocs completely new table  */
{
    table *new_tab;
    tuple_list *tpl;
    schema *sch = copy_schema(tb->sch);
    add_attribute_to_schema( sch, attrib );
    new_tab = mk_empty_table( sch );
    for ( tpl = tb->tups; tpl != NULL; tpl = tpl->next ) {
        tuple elem = mk_tuple( sch->size );
        int i;
        for ( i=0; i<sch->size-1; i++ ) {
            elem[i] = tpl->elem[i];
            }
        elem[i] = init;
        add_to_table( new_tab, elem );
        }
    return new_tab;
}

table *add_schema_to_table( table *tb, schema *new, tuple init )
               /* mallocs completely new table  */
{
    table *new_tab;
    tuple_list *tpl = tb->tups;
    schema *sch = copy_schema(tb->sch);
    int size = sch->size;
    add_schema_to_schema( sch, new  );
    new_tab = mk_empty_table( sch );
    for ( tpl = tb->tups; tpl != NULL; tpl = tpl->next ) {
        tuple elem = mk_tuple( sch->size );
        int i;
        for ( i=0; i<size; i++ ) {
            elem[i] = tpl->elem[i];
            }
        for ( i=0; i<new->size; i++ ) {
            elem[size+i] = init[i];
            }
        add_to_table( new_tab, elem );
        }
    return new_tab;
}

int first_currency( currency *cr, table *tb )
{
    cr->tb = tb;
    cr->pos = 0;
    cr->tup = tb->tups;
    return ( cr->tup != NULL );
}

int is_last_currency( currency *cr )
{
    return  ( cr->tup == NULL );
}

int next_currency( currency *cr )
{
    if ( cr->tup ) {
        cr->tup = cr->tup->next;
        cr->pos++;
        }
    return ( cr->tup != NULL );
}

tuple tuple_at_currency( currency *cr )
{
    if ( cr->tup ) return cr->tup->elem;
    else           return NULL;
}

int  set_num_in_currency ( currency *cr, attr_no attr, int num )
{
    schema *sch = cr->tb->sch;
    assert ( cr->tup != 0 );
    assert ( 0<=attr && attr < sch->size );
    if ( sch->elems[attr].type != ET_num ) return 0;
    cr->tup->elem[attr].num = num;
    return 1;
}

int  set_str_in_currency ( currency *cr, attr_no attr, char *str )
{
    schema *sch = cr->tb->sch;
    assert ( cr->tup != 0 );
    assert ( 0<=attr && attr < sch->size );
    if ( sch->elems[attr].type != ET_str ) return 0;
    cr->tup->elem[attr].str = str;
    return 1;
}

int  get_num_in_currency ( currency *cr, attr_no attr )
{
    schema *sch = cr->tb->sch;
    assert ( cr->tup != 0 );
    assert ( 0<=attr && attr < sch->size );
    assert ( sch->elems[attr].type == ET_num );
    return cr->tup->elem[attr].num;
}

char *get_str_in_currency ( currency *cr, attr_no attr )
{
    schema *sch = cr->tb->sch;
    assert ( cr->tup != 0 );
    assert ( 0<=attr && attr < sch->size );
    assert ( sch->elems[attr].type == ET_str );
    return cr->tup->elem[attr].str;
}

int  is_attr_num( attr_no attr, schema *sch )
{
    assert( 0 <= attr && attr < sch->size );
    return sch->elems[attr].type == ET_num;
}

int  is_attr_str( attr_no attr, schema *sch )
{
    assert( 0 <= attr && attr < sch->size );
    return sch->elems[attr].type == ET_str;
}

class_no get_class_no( char *name, class_sch *cl )
{
    return lookup(name,cl->nos_classes,cl->names);
}

char *name_of_class( class_no cn, class_sch *cl )
{
    assert( 0 <= cn && cn < cl->nos_classes );
    return cl->names[cn];
} 

attr_no get_attr_no( char *name, schema *sch )
{
    int i;
    for (i=0; i<sch->size; i++)
        if ( strcmp(name,sch->elems[i].name) == 0 ) return i;
    return -1;
}

char *name_of_attrib( attr_no attr, schema *sch )
{
    assert( 0 <= attr && attr < sch->size );
    return sch->elems[attr].name;
} 

