/**********************************/
/*                                */
/*     dl.c                       */
/*                                */
/*     decision list              */
/*     code                       */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     3/12/1993                  */
/*                                */
/**********************************/

#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "mem.h"

#include "dt.h"

static decision_list single_attr_decisions( attr_no attr, table *tab );
static decision_list pair_attr_decisions( attr_no attr, attr_no att2, table *tab );
static int find_next_smallest ( attr_no attr, int lo, table *tab );
static unsigned nice_pos_number_inbetween( unsigned lo, unsigned hi );
static int nice_number_inbetween( int lo, int hi );
static decision_list single_num_decisions( attr_no attr, table *tab );
static decision_list single_str_decisions( attr_no attr, table *tab );


decision_list append_dec_list( decision_list front, decision_list back )
{
    decision_list this;
    if ( front == NULL ) return back;
    for ( this=front; this->next; this=this->next ) /* do nout */;
    this->next = back;
    return front;
}

decision_list add_dec_list( decision_list dl, decision dec )
{
    decision_list new = (decision_list)mem_malloc(sizeof(struct dec_l_st));
    new->dec = dec;
    new->next = dl;
    return new;
}

void  free_dec_list( decision_list dl )
{
    decision_list this, next;
    this = dl;
    while ( this ) {
        next = this->next;
        mem_free(this);
        this = next;
        }
}


static decision_list single_attr_decisions( attr_no attr, table *tab )
{
    schema *sch = tab->sch;
    if ( is_attr_num(attr,sch) ) {
        return single_num_decisions( attr, tab );
        }
    else if ( is_attr_str(attr,sch) ){
        return single_str_decisions( attr, tab );
        }
    else return NULL;   /* other new data types */
}

static decision_list pair_attr_decisions( attr_no attr, attr_no att2, table *tab )
{
    schema *sch = tab->sch;
    decision_list  dl = NULL;
    decision dec;
    dec.attr = attr;
    dec.inf.pair.att2 = att2;
    
    if ( is_attr_num(attr,sch) && is_attr_num(att2,sch) ) {
        dec.class = N_pair;
        dec.op = OP_le;
        dl = add_dec_list( dl, dec );
        dec.op = OP_lt;
        dl = add_dec_list( dl, dec );
        dec.op = OP_eq;
        dl = add_dec_list( dl, dec );
        return dl;
        }
    else if ( is_attr_str(attr,sch) && is_attr_str(att2,sch) ){
        dec.class = S_pair;
        dec.op = OP_eq;
        dl = add_dec_list( dl, dec );
        return dl;
        }
    else return NULL;   /* other new data types */
}

static int find_next_smallest ( attr_no attr, int lo, table *tab )
      /* returns NUM_MAX when done */
{
    currency cur;
    int hi = NUM_MAX;
    for ( first_currency( &cur, tab );
                  !is_last_currency( &cur );
                          next_currency( &cur )  ) {
        int val = get_num_in_currency( &cur, attr );
        if ( val > lo && val < hi ) hi = val;
        }
    return hi;
}

static unsigned nice_pos_number_inbetween( unsigned lo, unsigned hi )
{
    char lo_buff[21], hi_buff[21];
    int i;

    sprintf(lo_buff,"%20d",lo);
    sprintf(hi_buff,"%20d",hi);
    i=0;
    while ( lo_buff[i] && lo_buff[i]==hi_buff[i] ) i++;
    if ( lo_buff[i] == '\0' )  return lo;   /*  equal  */
    if ( lo_buff[i] == ' ' ) hi_buff[i] = '1';
    else if ( lo_buff[i] < '5' && '5' <= hi_buff[i] )  hi_buff[i] = '5';
    else if ( lo_buff[i] < '2' && '2' <= hi_buff[i] )  hi_buff[i] = '2';
    for ( i++; hi_buff[i]; i++ ) hi_buff[i]='0';
    return atoi(hi_buff);
}

static int nice_number_inbetween( int lo, int hi )
{
    if ( lo < 0 && 0 < hi )  return 0;
    if ( lo < 0 )  return -nice_pos_number_inbetween( -hi, -lo );
    else           return  nice_pos_number_inbetween( lo, hi );
}

static decision_list single_num_decisions( attr_no attr, table *tab )
{
    int lo, hi, val;
    decision_list  dl = NULL ;
    decision dec;
    dec.attr = attr;
    dec.class = N_const;

    for ( lo = find_next_smallest ( attr, NUM_MIN, tab ),
          hi = find_next_smallest ( attr, lo, tab );
                  hi < NUM_MAX;
                          lo = hi,
                          hi = find_next_smallest ( attr, lo, tab ) ) {
        val = nice_number_inbetween(lo,hi);
        dec.inf.n_c.val = val;
        if ( val == hi ) {
            dec.op = OP_lt;
            dl = add_dec_list( dl, dec );
            dec.op = OP_eq;
            dl = add_dec_list( dl, dec );
            }
        else {
            dec.op = OP_le;
            dl = add_dec_list( dl, dec );
            }
        }
    return dl;
}

decision_list single_str_decisions( attr_no attr, table *tab )
{
    int lo, hi;
    decision_list  dl = NULL ;
    currency cur;
    decision dec;
    dec.attr = attr;
    dec.class = S_const;

    for ( first_currency( &cur, tab );
                  !is_last_currency( &cur );
                          next_currency( &cur )  ) {
        decision_list search;
        char *str = get_str_in_currency( &cur, attr );
        for ( search = dl; search; search = search->next ) {
            if ( strcmp(str,search->dec.inf.s_c.val) == 0 ) break;
            }
        if ( search ) continue;
        dec.inf.s_c.val = str;
        dec.op = OP_eq;
        dl = add_dec_list( dl, dec );
        }
    return dl;
}

decision_list all_possible_decisions( int nos_atts, attr_no *atts, table *tab )
{
    int i, j;
    decision_list dl = NULL, add;
    attr_no attr, att2;
    for ( i=0; i<nos_atts; i++ ) {
        attr = atts[i];
        add = single_attr_decisions( attr, tab );
        dl = append_dec_list( add, dl );
        }  
    for ( i=0; i<nos_atts; i++ ) {
        attr = atts[i];
        for ( j=i+1; j<nos_atts; j++ ) {
            att2 = atts[j];
            add = pair_attr_decisions( attr, att2, tab );
            dl = append_dec_list( add, dl );
        }   }
    return dl;
}
