/**********************************/
/*                                */
/*     dt.c                       */
/*                                */
/*     decision tree              */
/*     code                       */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "mem.h"

#include "dt.h"

int  classify( table *tab, decision_tree *dt, attr_no attr )
{
    currency cur;
    class_sch *cl = dt->cl;
    if ( first_currency( &cur, tab ) == 0 ) return 1;
    for ( ; ; ) {
        tuple tup = tuple_at_currency( &cur );
        class_no choice = classify_tuple( tup, dt );
        if ( choice < 0 ) return 0;
        set_str_in_currency( &cur, attr, name_of_class( choice, cl ));
        if ( next_currency( &cur ) == 0 ) return 1;
        }
}

table  *split_classify( table *tab, decision_tree *dt )
{
    currency cur;
    class_sch *cl = dt->cl;
    schema *sch = tab->sch;
    int n_cl = cl->nos_classes;
    table *split  = (table*)mem_malloc( n_cl * sizeof(table) );
    int i;

    assert(equal_schema(sch,dt->sch));

    for ( i=0; i<n_cl; i++ ) {
        schema *cpy = copy_schema(sch);
        empty_table(&split[i],cpy);
        }

    for ( first_currency( &cur, tab ) ;
                  ! is_last_currency( &cur) ;
                          next_currency( &cur )  ) {
        tuple tup = tuple_at_currency( &cur );
        class_no choice = classify_tuple( tup, dt );
        if ( choice < 0 ) return NULL;
        add_cur_to_table( &split[choice], &cur );
        }

   return split;
}

int  class_count( table *tab, decision_tree *dt, int *count )
         /*  count[nos_classes]  */
{
    currency cur;
    class_sch *cl = dt->cl;
    int i;
    for ( i=0; i<cl->nos_classes; i++ ) count[i] = 0;
    if ( first_currency( &cur, tab ) == 0 ) return 1;
    for ( ; ; ) {
        tuple tup = tuple_at_currency( &cur );
        class_no choice = classify_tuple( tup, dt );
        if ( choice < 0 ) return 0;
        count[choice]++;
        if ( next_currency( &cur ) == 0 ) return 1;
        }
}

class_no  classify_tuple( tuple tup, decision_tree *dt )
{
    schema    *sch  =  dt->sch;
    class_sch *cl   =  dt->cl;
    tree_node *tree =  dt->tree;

    while ( !is_leaf(tree->dec) ) {
        int choice = satisfies( tree->dec, tup, sch );
        if ( choice <0 || choice >= tree->nos_children ) return -1;
        tree = tree->children[choice];
        }
    return tree->dec.inf.c_c.choice;
}

/*
boolean may_choose( class_no choice, decision_tree *dt )
{
    schema    *sch  =  dt->sch;
    class_sch *cl   =  dt->cl;
    tree_node *tree =  dt->tree;
}
*/
 
 /**
  * returns true of the node is a leaf and chooses given class
  */   
int leaf_chooses( class_no choice, tree_node *tree )
{
    return is_leaf(tree->dec) && ( tree->dec.inf.c_c.choice == choice );
}
 /**
  * returns first child to exactly choose given class
  *         -1 if none
  */   
int child_for_class( class_no choice, tree_node *tree )
{
    if ( is_leaf(tree->dec) )
    {
    	return -1; // no children!
    }
    else
    {
    	int i;
    	for ( i=0; i<tree->nos_children; i++ )
    	{
    		if ( may_choose( choice, tree->children[i] ) ) return i;
    	}
    	return -1;
    }
}
 /**
  * returns whether any leaves choose given class
  */   
int may_choose( class_no choice, tree_node *tree )
{
    if ( is_leaf(tree->dec) )
    {
    	return ( tree->dec.inf.c_c.choice == choice );
    }
    else
    {
    	int i;
    	for ( i=0; i<tree->nos_children; i++ )
    	{
    		if ( may_choose( choice, tree->children[i] ) ) return 1;
    	}
    	return 0;
    }
}
 /**
  * returns number of leaves that can choose given class
  */   
int nos_leaves_to_choose( class_no choice, tree_node *tree )
{
    if ( is_leaf(tree->dec) )
    {
    	if ( tree->dec.inf.c_c.choice == choice )
    		return 1;
    	else
    		return 0;
    }
    else
    {
    	int i;
    	int sum = 0;
    	for ( i=0; i<tree->nos_children; i++ )
    	{
    		sum += nos_leaves_to_choose( choice, tree->children[i] );
    	}
    	return sum;
    }
}

 /**
  * returns number of top level branches that could choose given class
  */   
int nos_branches_to_choose( class_no choice, tree_node *tree )
{
    if ( is_leaf(tree->dec) )
    {
    	if ( tree->dec.inf.c_c.choice == choice )
    		return 1;
    	else
    		return 0;
    }
    else
    {
    	int i;
    	int sum = 0;
    	for ( i=0; i<tree->nos_children; i++ )
    	{
    	    if ( may_choose( choice, tree->children[i] ) ) sum++;
    	}
    	return sum;
    }
}


int satisfies_cur( decision dec, currency *cur )
                /*  N.B. for binary choices 0 = YES, 1 = NO  */
{
    schema *sch = cur->tb->sch;
    tuple tup = tuple_at_currency( cur );
    return satisfies( dec, tup, sch );
}


int satisfies( decision dec, tuple tup, schema *sch )
                /*  N.B. for binary choices 0 = YES, 1 = NO  */
{
    int nv, nv2;
    char *sv, *sv2;
    char **ssv2;

    switch ( dec.class ) {
        case C_leaf:    return 0;
        case N_const: case N_pair:  
                        nv = tup[dec.attr].num;
                        if ( dec.class == N_const )
                            nv2 = dec.inf.n_c.val;
                        else
                            nv2 = tup[dec.inf.pair.att2].num;
                        if ( dec.op == OP_eq )
                              return !( nv == nv2 );
                        else if ( dec.op == OP_le )
                              return !( nv <= nv2 );
                        else  return !( nv <  nv2 );
        case S_const: case S_pair:
                        sv  = tup[dec.attr].str;
                        if ( dec.class == S_const )
                            sv2 = dec.inf.s_c.val;
                        else
                            sv2 = tup[dec.inf.pair.att2].str;
                        return !( strcmp(sv,sv2) == 0 );
        case S_set:     return 0;   /* not yet implmented */ 
        default:        return 0;
      }
}

