/**********************************/
/*                                */
/*     dt_io.c                    */
/*                                */
/*     decision tree              */
/*     input/output               */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "mem.h"

#include "gen_io.h"
#include "dt.h"
#include "dt.h"

typedef enum and_or_state_enum { aos_top=0, aos_and=1, aos_or=2 } and_or_state;

static void print_tn( FILE *fp, tree_node *tn, decision_tree *dt, int indent );
static void print_decision( FILE *fp, decision dec, decision_tree *dt );
static void print_tn_as_and_or( FILE *fp, class_no c, tree_node *tn, decision_tree *dt, and_or_state state );
static void print_branch_as_and_or( FILE *fp, int child_idx, class_no c, tree_node *tn, decision_tree *dt, and_or_state state );
static void print_decision_as_boolean( FILE *fp, decision dec, decision_tree *dt, int choice );

static int read_tn( FILE *fp, tree_node **tn, decision_tree *dt );
static int read_decision( FILE *fp, decision *dec, decision_tree *dt );
static int read_if_decision( FILE *fp, decision *dec, decision_tree *dt );
static int read_class_decision( FILE *fp, decision *dec, decision_tree *dt );
static int read_if_str_in_set( FILE *fp, decision *dec, decision_tree *dt );
static int read_if_str( FILE *fp, decision *dec, decision_tree *dt );
static int read_if_num( FILE *fp, decision *dec, decision_tree *dt );
static int read_operator( FILE *fp, dec_op *op, decision_tree *dt );

void print_decision_tree( FILE *fp, decision_tree *dt )
{
    print_tn(fp,dt->tree,dt,0);
}

void print_decision_tree_as_and_or( FILE *fp, class_no choice, decision_tree *dt )
{
    print_tn_as_and_or(fp,choice,dt->tree,dt,aos_top);
}

int read_decision_tree( FILE *fp, decision_tree *dt, schema *sch, class_sch *cl )
{
    int ret, c;
    dt->sch = sch;
    dt->cl  = cl;
    ret = read_tn(fp,&(dt->tree),dt);
    if ( ret ) return 1;
    print_read_error(stderr,fp,"reading tree");
    return 0;
}


static void print_tn( FILE *fp, tree_node *tn, decision_tree *dt, int indent )
{
    int ch;
    print_indent( fp, indent );
    print_decision( fp, tn->dec, dt );
    fputc('\n',fp);
    if ( is_leaf( tn->dec ) ) return;
    for ( ch=0; ch<tn->nos_children; ch++ ) {
        print_tn(fp,tn->children[ch],dt,indent+4);
        }
}

/**
 *     change tree:
 *       IF cond0
 *           CASE 1 IF cond1
 *              CASE 11 CLASS x
 *              CASE 12 IF cond12
 *                  CASE 121 CLASS y
 *                  CASE 122 CLASS x
 *              CASE 13 CLASS y
 *           CASE 2 CLASS y
 *           CASE 3 IF cond3
 *              CASE 31 IF cond31
 *                  CASE 311 IF cond311
 *                      CASE 3111 CLASS y
 *                      CASE 3112 CLASS x
 *                  CASE 312 CLASS x
 *              CASE 32 CLASS x
 *              CASE 33 CLASS y
 *
 *      into (for class y):
 *        ( cond0is1 AND ( (cond1is2 AND cond12is1) OR cond1is3 ) ) OR
 *        cond0is2 OR
 *        ( cond0is3 AND ( (cond3is1 AND cond31is1 AND cond311is1) OR cond3is3 ) )
 * 
 *     rules:
 *       if there are multiple paths to given class need OR between them
 *       if OR under AND need brackets around them
 *       if next level down 
 *
 */

static void print_tn_as_and_or( FILE *fp, class_no choice, tree_node *tn, decision_tree *dt, and_or_state state )
{
	int nos_of_ways;
	
    if ( is_leaf( tn->dec ) )  // shouldn't really happen except at top level
    {
    	if ( tn->dec.inf.c_c.choice == choice )
    		if ( state==aos_top ) fprintf(fp,"TRUE");
    	else
    		if ( state==aos_top ) fprintf(fp,"FALSE");
    	return;
    }

    nos_of_ways = nos_branches_to_choose( choice, tn );
    
    if ( nos_of_ways==0 )  // shouldn't happen except at top level
    {
    	if ( state==aos_top ) fprintf(fp,"FALSE");
    	return;
    }
    else if ( nos_of_ways == 1 )
    {
    	int child_idx = child_for_class( choice, tn );
    	print_branch_as_and_or(fp, child_idx, choice, tn, dt, state);
    }
    else // nos_of_ways > 1
    {
    	int needs_brackets = ( state==aos_and );
    	int i;
   		int first = 1;
    	if ( needs_brackets ) fprintf(fp,"( ");
    	for ( i=0; i<tn->nos_children && i<2; i++ )  // should normally be exactly 2
    	{
    		tree_node *child = tn->children[i];
    		if ( ! may_choose( choice, child ) ) continue;
    		if ( first ) first=0;
    		else fprintf(fp," OR ");
    		print_branch_as_and_or(fp, i, choice, tn, dt, aos_or);
    	}
    	if ( needs_brackets ) fprintf(fp," )");
    }
}


static void print_branch_as_and_or( FILE *fp, int child_idx, class_no choice, tree_node *tn, decision_tree *dt, and_or_state state )
{
    tree_node *child = tn->children[child_idx];
    int only_choice = leaf_chooses(  choice, child );
    int needs_brackets = ( state == aos_or && !only_choice );
    if ( needs_brackets ) fprintf(fp,"( ");
    print_decision_as_boolean(fp, tn->dec, dt, child_idx);
    if ( !only_choice )
    {
    	fprintf(fp," AND ");
    	print_tn_as_and_or(fp, choice, child, dt, aos_and);
    }
    if ( needs_brackets ) fprintf(fp," )");
}

static int read_tn( FILE *fp, tree_node **tn, decision_tree *dt )
{
    decision dec;
    int ret, ch, nc;
    ret = read_decision( fp, &dec, dt );
    if ( !ret ) return 0;
    *tn = (tree_node*)mem_malloc(sizeof(tree_node));
    (*tn)->dec = dec;
    if ( is_leaf( dec ) ) {
        (*tn)->nos_children = 0;
        return 1;
        }

    nc = nos_choices( dec );

    (*tn)->nos_children = nc;
    (*tn)->children = (tree_node**)mem_malloc( nc * sizeof(tree_node*) );

    for ( ch=0; ch<nc; ch++ ) {
        ret = read_tn(fp,&((*tn)->children[ch]),dt);
        if ( !ret ) return 0;
        }
}

void print_dec_list( FILE *fp, decision_list dl, decision_tree *dt )
{
    decision_list this;
    for ( this=dl; this; this=this->next ) {
        print_decision( fp, this->dec, dt );
        fputc('\n',fp);
        }
}


static void print_decision( FILE *fp, decision dec, decision_tree *dt )
{
    char *attr_name, *att2_name, *op_name;
    if ( dec.class == C_leaf ) {
        char *class_name = name_of_class(dec.inf.c_c.choice,dt->cl);
        int cf = dec.inf.c_c.cf;
        fprintf(fp,"CLASSIFIED %s",class_name);
        if ( cf < 100 ) fprintf(fp,"  @ %d%%",cf);
        return;
        }
    attr_name = name_of_attrib(dec.attr,dt->sch);
    op_name = name_of_dec_op( dec.op );
    fprintf(fp,"IF %s %s ",attr_name,op_name);
    switch ( dec.class ) {
        case N_const: fprintf(fp,"%d",dec.inf.n_c.val);
                      break;
        case S_const: fprintf(fp,"\"%s\"",dec.inf.s_c.val);
                      break;
        case S_set:   fprintf(fp,"{}");    /*  not implemented yet  */
                      break;
        case N_pair: case S_pair:
                      att2_name = name_of_attrib(dec.inf.pair.att2,dt->sch);
                      fprintf(fp,"%s",att2_name);
                      break;
        }
}

static void print_decision_as_boolean( FILE *fp, decision dec, decision_tree *dt, int choice )
{
    char *attr_name, *att2_name, *op_name;
    if ( dec.class == C_leaf ) { // shouldn't happen
        fprintf(fp,"ERROR formating classification as decision");
        return;
    }
    attr_name = name_of_attrib(dec.attr,dt->sch);
    if ( choice == 0 )
    	op_name = name_of_dec_op( dec.op );
    else
    	op_name = name_of_not_dec_op( dec.op );
    fprintf(fp,"%s %s ",attr_name,op_name);
    switch ( dec.class ) {
        case N_const: fprintf(fp,"%d",dec.inf.n_c.val);
                      break;
        case S_const: fprintf(fp,"\"%s\"",dec.inf.s_c.val);
                      break;
        case S_set:   fprintf(fp,"{}");    /*  not implemented yet  */
                      break;
        case N_pair: case S_pair:
                      att2_name = name_of_attrib(dec.inf.pair.att2,dt->sch);
                      fprintf(fp,"%s",att2_name);
                      break;
        }
}

static int read_decision( FILE *fp, decision *dec, decision_tree *dt )
{
    token my_tok = read_token(fp);
    if ( tok_is_var("IF") )
            return read_if_decision( fp, dec, dt );
    if ( tok_is_var("CLASSIFIED") )
            return read_class_decision( fp, dec, dt );
    read_error = "expecting IF or CLASSIFIED";
    return 0;
}

static int read_class_decision( FILE *fp, decision *dec, decision_tree *dt )
{
    token my_tok = read_token(fp);
    class_no choice;
    int c, cf;
    if ( my_tok != TOK_VAR ) {
        read_error = "expecting class name";
        return 0;
        }
    choice = get_class_no(tok_val,dt->cl);
    if ( choice < 0 ) {
        read_error = "bad class name";
        return 0;
        }
    c = fgetc(fp);
    if ( c != '@' ) {
        ungetc(c,fp);
        dec->class = C_leaf;
        dec->op    = OP_none;
        dec->inf.c_c.choice = choice;
        dec->inf.c_c.cf     = 100;
        return 1;
        }
    my_tok = read_token(fp);
    if ( my_tok != TOK_NUM ) {
        read_error = "expecting certainty factor";
        return 0;
        }
    cf = tok_num;
    c = fgetc(fp);
    if ( c != '%' ) ungetc(c,fp);
    if ( cf > 100 ) {
        read_error = "certainty factor > 100%";
        return 0;
        }
    dec->class = C_leaf;
    dec->op    = OP_none;
    dec->inf.c_c.choice = choice;
    dec->inf.c_c.cf     = cf;
    return 1;
}


static int read_if_decision( FILE *fp, decision *dec, decision_tree *dt )
{
    token my_tok = read_token(fp);
    attr_no attr;
    int ret;

    if ( my_tok != TOK_VAR ) {
        read_error = "expecting attribute name";
        return 0;
        }
    attr = get_attr_no(tok_val,dt->sch);
    if ( attr < 0 ) {
        read_error = "bad attribute name";
        return 0;
        }
    dec->attr = attr;
    ret = read_operator( fp, &(dec->op), dt );
    if ( !ret ) return 0;
    if ( is_attr_num(attr,dt->sch) )
            return read_if_num( fp, dec, dt );
    else if ( is_attr_str(attr,dt->sch) ) {
       if ( dec->op == OP_in )
            return read_if_str_in_set( fp, dec, dt );
       else return read_if_str( fp, dec, dt );
       }
    assert(0);
    read_error = "*** INTERNAL ERROR ***";
    return 0;
}

static int read_operator( FILE *fp, dec_op *op, decision_tree *dt )
{
    token my_tok = read_token(fp);
    if ( my_tok == TOK_SYM || my_tok == TOK_VAR ) {
        *op = lookup_dec_op(tok_val);
        if ( *op != OP_none )  return 1;
        }
    read_error = "expecting operator";
    return 0;
}

static int read_if_num( FILE *fp, decision *dec, decision_tree *dt )
{
    token my_tok;
    int  val;
    attr_no attr;

    if ( dec->op!=OP_eq && dec->op!=OP_le && dec->op!=OP_lt ) {
        read_error = "bad operator for number";
        return 0;
        }

    my_tok = read_token(fp);

    if ( tok_is_num() ) {
        dec->class = N_const;
        dec->inf.n_c.val = tok_num;
        return 1;
        }
    if ( my_tok != TOK_VAR ) {
        read_error = "expecting number or attribute name";
        return 0;
        }
    attr = get_attr_no(tok_val,dt->sch);
    if ( attr < 0 ) {
        read_error = "bad attribute name";
        return 0;
        }
    if ( is_attr_num(attr,dt->sch) ) {
        dec->class = N_pair;
        dec->inf.pair.att2 = attr;
        return 1;
        }
    read_error = "can't compare number with string";
    return 0;
}

static int read_if_str_in_set( FILE *fp, decision *dec, decision_tree *dt )
{
    token my_tok;
    dec->class = S_set;
    dec->inf.s_s.nos_elems = 0;
    dec->inf.s_s.elems = NULL;

    my_tok = read_token(fp);
    if ( ! tok_is_sym("{") ) {
        read_error = "set must begin with }";
        return 0;
        }

    do {
        my_tok = read_token(fp);
       } while ( ! tok_is_sym("}") && ! tok_is_eof() );

    if ( ! tok_is_sym("}") ) {
        read_error = "set must end with }";
        return 0;
        }
    
    return 1;

}


static int read_if_str( FILE *fp, decision *dec, decision_tree *dt )
{
    token my_tok;
    int  val;
    attr_no attr;

    if ( dec->op!=OP_eq ) {
        read_error = "bad operator for string";
        return 0;
        }

    my_tok = read_token(fp);

    if ( tok_is_str() ) {
        char *val = (char*)mem_malloc(tok_len+1);
        strcpy( val, tok_val );
        dec->class = S_const;
        dec->inf.s_c.val = val;
        return 1;
        }
    if ( my_tok != TOK_VAR ) {
        read_error = "expecting literal string or attribute name";
        return 0;
        }
    attr = get_attr_no(tok_val,dt->sch);
    if ( attr < 0 ) {
        read_error = "bad attribute name";
        return 0;
        }
    if ( is_attr_str(attr,dt->sch) ) {
        dec->class = S_pair;
        dec->inf.pair.att2 = attr;
        return 1;
        }
    read_error = "can't compare number with string";
    return 0;
}



